"""Provide mashumaro data object for AirOSData."""

from dataclasses import dataclass, field
from enum import Enum
from typing import Any

from mashumaro import DataClassDictMixin


class IeeeMode(Enum):
    """Enum definition."""

    AUTO = "AUTO"
    _11ACVHT80 = "11ACVHT80"
    # More to be added when known


class WirelessMode(Enum):
    """Enum definition."""

    AccessPoint_PointToPoint = "ap-ptp"
    Station_PointToPoint = "sta-ptp"
    # More to be added when known


class Security(Enum):
    """Enum definition."""

    WPA2 = "WPA2"
    # More to be added when known


class NetRole(Enum):
    """Enum definition."""

    BRIDGE = "bridge"
    ROUTER = "router"
    # More to be added when known


@dataclass
class ChainName:
    """Leaf definition."""

    number: int
    name: str


@dataclass
class Host:
    """Leaf definition."""

    hostname: str
    device_id: str
    uptime: int
    power_time: int
    time: str
    timestamp: int
    fwversion: str
    devmodel: str
    netrole: NetRole | str
    loadavg: float
    totalram: int
    freeram: int
    temperature: int
    cpuload: float
    height: int


@dataclass
class Services:
    """Leaf definition."""

    dhcpc: bool
    dhcpd: bool
    dhcp6d_stateful: bool
    pppoe: bool
    airview: int


@dataclass
class Firewall:
    """Leaf definition."""

    iptables: bool
    ebtables: bool
    ip6tables: bool
    eb6tables: bool


@dataclass
class Throughput:
    """Leaf definition."""

    tx: int
    rx: int


@dataclass
class ServiceTime:
    """Leaf definition."""

    time: int
    link: int


@dataclass
class Polling:
    """Leaf definition."""

    cb_capacity: int
    dl_capacity: int
    ul_capacity: int
    use: int
    tx_use: int
    rx_use: int
    atpc_status: int
    fixed_frame: bool
    gps_sync: bool
    ff_cap_rep: bool


@dataclass
class Stats:
    """Leaf definition."""

    rx_bytes: int
    rx_packets: int
    rx_pps: int
    tx_bytes: int
    tx_packets: int
    tx_pps: int


@dataclass
class EvmData:
    """Leaf definition."""

    usage: int
    cinr: int
    evm: list[list[int]]


@dataclass
class Airmax:
    """Leaf definition."""

    actual_priority: int
    beam: int
    desired_priority: int
    cb_capacity: int
    dl_capacity: int
    ul_capacity: int
    atpc_status: int
    rx: EvmData
    tx: EvmData


@dataclass
class EthList:
    """Leaf definition."""

    ifname: str
    enabled: bool
    plugged: bool
    duplex: bool
    speed: int
    snr: list[int]
    cable_len: int


@dataclass
class GPSData:
    """Leaf definition."""

    lat: str
    lon: str
    fix: int


@dataclass
class Remote:
    """Leaf definition."""

    age: int
    device_id: str
    hostname: str
    platform: str
    version: str
    time: str
    cpuload: float
    temperature: int
    totalram: int
    freeram: int
    netrole: str
    mode: WirelessMode | str  # Allow non-breaking future expansion
    sys_id: str
    tx_throughput: int
    rx_throughput: int
    uptime: int
    power_time: int
    compat_11n: int
    signal: int
    rssi: int
    noisefloor: int
    tx_power: int
    distance: int
    rx_chainmask: int
    chainrssi: list[int]
    tx_ratedata: list[int]
    tx_bytes: int
    rx_bytes: int
    antenna_gain: int
    cable_loss: int
    height: int
    ethlist: list[EthList]
    ipaddr: list[str]
    ip6addr: list[str]
    gps: GPSData
    oob: bool
    unms: dict[str, Any]
    airview: int
    service: ServiceTime


@dataclass
class Station:
    """Leaf definition."""

    mac: str
    lastip: str
    signal: int
    rssi: int
    noisefloor: int
    chainrssi: list[int]
    tx_idx: int
    rx_idx: int
    tx_nss: int
    rx_nss: int
    tx_latency: int
    distance: int
    tx_packets: int
    tx_lretries: int
    tx_sretries: int
    uptime: int
    dl_signal_expect: int
    ul_signal_expect: int
    cb_capacity_expect: int
    dl_capacity_expect: int
    ul_capacity_expect: int
    dl_rate_expect: int
    ul_rate_expect: int
    dl_linkscore: int
    ul_linkscore: int
    dl_avg_linkscore: int
    ul_avg_linkscore: int
    tx_ratedata: list[int]
    stats: Stats
    airmax: Airmax
    last_disc: int
    remote: Remote


@dataclass
class Wireless:
    """Leaf definition."""

    essid: str
    mode: WirelessMode | str  # Allow non-breaking expansion
    ieeemode: IeeeMode | str  # Allow non-breaking expansion
    band: int
    compat_11n: int
    hide_essid: int
    apmac: str
    antenna_gain: int
    frequency: int
    center1_freq: int
    dfs: int
    distance: int
    security: Security | str  # Allow non-breaking expansion
    noisef: int
    txpower: int
    aprepeater: bool
    rstatus: int
    chanbw: int
    rx_chainmask: int
    tx_chainmask: int
    nol_state: int
    nol_timeout: int
    cac_state: int
    cac_timeout: int
    rx_idx: int
    rx_nss: int
    tx_idx: int
    tx_nss: int
    throughput: Throughput
    service: ServiceTime
    polling: Polling
    count: int
    sta: list[Station]
    sta_disconnected: list[Any]


@dataclass
class InterfaceStatus:
    """Leaf definition."""

    plugged: bool
    tx_bytes: int
    rx_bytes: int
    tx_packets: int
    rx_packets: int
    tx_errors: int
    rx_errors: int
    tx_dropped: int
    rx_dropped: int
    ipaddr: str
    speed: int
    duplex: bool
    snr: list[int] | None = None
    cable_len: int | None = None
    ip6addr: list[dict[str, Any]] | None = None


@dataclass
class Interface:
    """Leaf definition."""

    ifname: str
    hwaddr: str
    enabled: bool
    mtu: int
    status: InterfaceStatus


@dataclass
class ProvisioningMode:
    """Leaf definition."""

    pass


@dataclass
class NtpClient:
    """Leaf definition."""

    pass


@dataclass
class UnmsStatus:
    """Leaf definition."""

    status: int


@dataclass
class GPSMain:
    """Leaf definition."""

    lat: float
    lon: float
    fix: int


@dataclass
class AirOSData(DataClassDictMixin):
    """Dataclass for AirOS devices."""

    chain_names: list[ChainName]
    host: Host
    genuine: str
    services: Services
    firewall: Firewall
    portfw: bool
    wireless: Wireless
    interfaces: list[Interface]
    provmode: (
        ProvisioningMode | str | dict[str, Any] | list[Any] | Any
    )  # If it can be populated, define its fields
    ntpclient: (
        NtpClient | str | dict[str, Any] | list[Any] | Any
    )  # If it can be populated, define its fields
    unms: UnmsStatus
    gps: GPSMain
    warnings: dict[str, list[str]] = field(default_factory=dict, init=False)

    @classmethod
    def __post_deserialize__(cls, airos_object: "AirOSData") -> "AirOSData":
        """Validate after deserialization."""
        airos_object.check_for_warnings()
        return airos_object

    def check_for_warnings(self):
        """Validate unions for unknown fields."""
        # Check wireless mode
        if isinstance(self.wireless.mode, str):
            self.add_warning(
                "wireless", f"Unknown (new) wireless mode: '{self.wireless.mode}'"
            )

        # Check host netrole
        if isinstance(self.host.netrole, str):
            self.add_warning(
                "host", f"Unknown (new) network role: '{self.host.netrole}'"
            )

        # Check wireless IEEE mode
        if isinstance(self.wireless.ieeemode, str):
            self.add_warning(
                "wireless", f"Unknown (new) IEEE mode: '{self.wireless.ieeemode}'"
            )

        # Check wireless security
        if isinstance(self.wireless.security, str):
            self.add_warning(
                "wireless", f"Unknown (new) security type: '{self.wireless.security}'"
            )
        # Check station remote modes
        for i, station in enumerate(self.wireless.sta):
            if hasattr(station.remote, "mode") and isinstance(station.remote.mode, str):
                self.add_warning(
                    f"wireless.sta[{i}].remote",
                    f"Unknown (new) remote mode: '{station.remote.mode}', please report to the CODEOWNERS for inclusion",
                )

    def add_warning(self, field_name: str, message: str):
        """Insert warnings into the dictionary on unknown field data."""
        if field_name not in self.warnings:
            self.warnings[field_name] = []
        self.warnings[field_name].append(message)



"""Ubiquiti AirOS 8 module for Home Assistant Core."""

from __future__ import annotations

import json
import logging
from urllib.parse import urlparse

import aiohttp
from mashumaro.exceptions import InvalidFieldValue, MissingField

from .airos8data import AirOSData
from .exceptions import (
    ConnectionAuthenticationError,
    ConnectionSetupError,
    DataMissingError,
    DeviceConnectionError,
    KeyDataMissingError,
)

logger = logging.getLogger(__name__)


class AirOS:
    """Set up connection to AirOS."""

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        session: aiohttp.ClientSession,
        use_ssl: bool = True,
    ):
        """Initialize AirOS8 class."""
        self.username = username
        self.password = password

        parsed_host = urlparse(host)
        scheme = (
            parsed_host.scheme
            if parsed_host.scheme
            else ("https" if use_ssl else "http")
        )
        hostname = parsed_host.hostname if parsed_host.hostname else host

        self.base_url = f"{scheme}://{hostname}"

        self.session = session

        self._login_url = f"{self.base_url}/api/auth"  # AirOS 8
        self._status_cgi_url = f"{self.base_url}/status.cgi"  # AirOS 8
        self._stakick_cgi_url = f"{self.base_url}/stakick.cgi"  # AirOS 8
        self.current_csrf_token = None
        self.warnings_cache = []

        self._use_json_for_login_post = False

        self._common_headers = {
            "Accept": "application/json, text/javascript, */*; q=0.01",
            "Sec-Fetch-Site": "same-origin",
            "Accept-Language": "en-US,nl;q=0.9",
            "Accept-Encoding": "gzip, deflate, br",
            "Sec-Fetch-Mode": "cors",
            "Origin": self.base_url,
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.5 Safari/605.1.15",
            "Referer": self.base_url + "/",
            "Connection": "keep-alive",
            "Sec-Fetch-Dest": "empty",
            "X-Requested-With": "XMLHttpRequest",
        }

        self.connected = False

    async def login(self) -> bool:
        """Log in to the device assuring cookies and tokens set correctly."""
        # --- Step 0: Pre-inject the 'ok=1' cookie before login POST (mimics curl) ---
        self.session.cookie_jar.update_cookies({"ok": "1"})

        # --- Step 1: Attempt Login to /api/auth (This now sets all session cookies and the CSRF token) ---
        login_payload = {
            "username": self.username,
            "password": self.password,
        }

        login_request_headers = {**self._common_headers}

        post_data = None
        if self._use_json_for_login_post:
            login_request_headers["Content-Type"] = "application/json"
            post_data = json.dumps(login_payload)
        else:
            login_request_headers["Content-Type"] = (
                "application/x-www-form-urlencoded; charset=UTF-8"
            )
            post_data = login_payload

        try:
            async with self.session.post(
                self._login_url,
                data=post_data,
                headers=login_request_headers,
            ) as response:
                if not response.cookies:
                    logger.exception("Empty cookies after login, bailing out.")
                    raise ConnectionSetupError from None
                else:
                    for _, morsel in response.cookies.items():
                        # If the AIROS_ cookie was parsed but isn't automatically added to the jar, add it manually
                        if (
                            morsel.key.startswith("AIROS_")
                            and morsel.key not in self.session.cookie_jar
                        ):
                            # `SimpleCookie`'s Morsel objects are designed to be compatible with cookie jars.
                            # We need to set the domain if it's missing, otherwise the cookie might not be sent.
                            # For IP addresses, the domain is typically blank.
                            # aiohttp's jar should handle it, but for explicit control:
                            if not morsel.get("domain"):
                                morsel["domain"] = (
                                    response.url.host
                                )  # Set to the host that issued it
                            self.session.cookie_jar.update_cookies(
                                {
                                    morsel.key: morsel.output(header="")[
                                        len(morsel.key) + 1 :
                                    ]
                                    .split(";")[0]
                                    .strip()
                                },
                                response.url,
                            )
                            # The update_cookies method can take a SimpleCookie morsel directly or a dict.
                            # The morsel.output method gives 'NAME=VALUE; Path=...; HttpOnly'
                            # We just need 'NAME=VALUE' or the morsel object itself.
                            # Let's use the morsel directly which is more robust.
                            # Alternatively: self.session.cookie_jar.update_cookies({morsel.key: morsel.value}) might work if it's simpler.
                            # Aiohttp's update_cookies takes a dict mapping name to value.
                            # To pass the full morsel with its attributes, we need to add it to the jar's internal structure.
                            # Simpler: just ensure the key-value pair is there for simple jar.

                            # Let's try the direct update of the key-value
                            self.session.cookie_jar.update_cookies(
                                {morsel.key: morsel.value}
                            )

                new_csrf_token = response.headers.get("X-CSRF-ID")
                if new_csrf_token:
                    self.current_csrf_token = new_csrf_token
                else:
                    return

                # Re-check cookies in self.session.cookie_jar AFTER potential manual injection
                airos_cookie_found = False
                ok_cookie_found = False
                if not self.session.cookie_jar:  # pragma: no cover
                    logger.exception(
                        "COOKIE JAR IS EMPTY after login POST. This is a major issue."
                    )
                    raise ConnectionSetupError from None
                for cookie in self.session.cookie_jar:  # pragma: no cover
                    if cookie.key.startswith("AIROS_"):
                        airos_cookie_found = True
                    if cookie.key == "ok":
                        ok_cookie_found = True

                if not airos_cookie_found and not ok_cookie_found:
                    raise ConnectionSetupError from None  # pragma: no cover

                response_text = await response.text()

                if response.status == 200:
                    try:
                        json.loads(response_text)
                        self.connected = True
                        return True
                    except json.JSONDecodeError as err:
                        logger.exception("JSON Decode Error")
                        raise DataMissingError from err

                else:
                    log = f"Login failed with status {response.status}. Full Response: {response.text}"
                    logger.error(log)
                    raise ConnectionAuthenticationError from None
        except (
            aiohttp.ClientError,
            aiohttp.client_exceptions.ConnectionTimeoutError,
        ) as err:
            logger.exception("Error during login")
            raise DeviceConnectionError from err

    async def status(self) -> AirOSData:
        """Retrieve status from the device."""
        if not self.connected:
            logger.error("Not connected, login first")
            raise DeviceConnectionError from None

        # --- Step 2: Verify authenticated access by fetching status.cgi ---
        authenticated_get_headers = {**self._common_headers}
        if self.current_csrf_token:
            authenticated_get_headers["X-CSRF-ID"] = self.current_csrf_token

        try:
            async with self.session.get(
                self._status_cgi_url,
                headers=authenticated_get_headers,
            ) as response:
                if response.status == 200:
                    try:
                        response_text = await response.text()
                        response_json = json.loads(response_text)
                        try:
                            airos_data = AirOSData.from_dict(response_json)
                        except (MissingField, InvalidFieldValue) as err:
                            logger.exception("Failed to deserialize AirOS data")
                            raise KeyDataMissingError from err

                        # Show new enums detected, once after (each) startup
                        if airos_data.warnings:
                            for field_name, messages in airos_data.warnings.items():
                                for msg in messages:
                                    log = f"AirOS data warning for field '{field_name}': {msg}"
                                    if log not in self.warnings_cache:
                                        self.warnings_cache.append(log)
                                        logger.warning(log)

                        return airos_data
                    except json.JSONDecodeError:
                        logger.exception(
                            "JSON Decode Error in authenticated status response"
                        )
                        raise DataMissingError from None
                else:
                    log = f"Authenticated status.cgi failed: {response.status}. Response: {response_text}"
                    logger.error(log)
        except (
            aiohttp.ClientError,
            aiohttp.client_exceptions.ConnectionTimeoutError,
        ) as err:
            logger.exception("Error during authenticated status.cgi call")
            raise DeviceConnectionError from err

    async def stakick(self, mac_address: str = None) -> bool:
        """Reconnect client station."""
        if not self.connected:
            logger.error("Not connected, login first")
            raise DeviceConnectionError from None
        if not mac_address:
            logger.error("Device mac-address missing")
            raise DataMissingError from None

        kick_request_headers = {**self._common_headers}
        if self.current_csrf_token:
            kick_request_headers["X-CSRF-ID"] = self.current_csrf_token

        kick_payload = {"staif": "ath0", "staid": mac_address.upper()}

        kick_request_headers["Content-Type"] = (
            "application/x-www-form-urlencoded; charset=UTF-8"
        )
        post_data = kick_payload

        try:
            async with self.session.post(
                self._stakick_cgi_url,
                headers=kick_request_headers,
                data=post_data,
            ) as response:
                if response.status == 200:
                    return True
                response_text = await response.text()
                log = f"Unable to restart connection response status {response.status} with {response_text}"
                logger.error(log)
                return False
        except (
            aiohttp.ClientError,
            aiohttp.client_exceptions.ConnectionTimeoutError,
        ) as err:
            logger.exception("Error during reconnect stakick.cgi call")
            raise DeviceConnectionError from err
